@@ -4,21 +4,16 @@ from __future__ import annotations
 
 import asyncio
 import logging
-from collections.abc import Iterable
-from copy import deepcopy
 from dataclasses import dataclass
 from datetime import datetime
 from typing import TYPE_CHECKING, Any, cast
 
 from aiohttp.client_exceptions import ServerDisconnectedError
-
-try:
-    from pydantic.v1 import PrivateAttr, ValidationError
-except ImportError:
-    from pydantic import PrivateAttr, ValidationError  # type: ignore[assignment]
+from convertertools import pop_dict_set, pop_dict_tuple
+from pydantic import PrivateAttr, ValidationError
 
 from ..exceptions import ClientError
-from ..utils import normalize_mac, utc_now
+from ..utils import normalize_mac, to_snake_case, utc_now
 from .base import (
     RECENT_EVENT_MAX,
     ProtectBaseObject,
@@ -26,20 +21,22 @@ from .base import (
     ProtectModel,
     ProtectModelWithId,
 )
-from .convert import create_from_unifi_dict
+from .convert import MODEL_TO_CLASS, create_from_unifi_dict
 from .devices import (
+    AiPort,
     Bridge,
     Camera,
     Chime,
     Doorlock,
     Light,
     ProtectAdoptableDeviceModel,
+    Ringtone,
     Sensor,
     Viewer,
 )
 from .nvr import NVR, Event, Liveview
 from .types import EventType, FixSizeOrderedDict, ModelType
-from .user import Group, User
+from .user import Group, Keyrings, UlpUserKeyringBase, UlpUsers, User
 from .websocket import (
     WSAction,
     WSPacket,
@@ -55,6 +52,7 @@ _LOGGER = logging.getLogger(__name__)
 MAX_SUPPORTED_CAMERAS = 256
 MAX_EVENT_HISTORY_IN_STATE_MACHINE = MAX_SUPPORTED_CAMERAS * 2
 STATS_KEYS = {
+    "eventStats",
     "storageStats",
     "stats",
     "systemInfo",
@@ -65,9 +63,34 @@ STATS_KEYS = {
     "lastSeen",
     "recordingSchedules",
 }
+
 IGNORE_DEVICE_KEYS = {"nvrMac", "guid"}
 STATS_AND_IGNORE_DEVICE_KEYS = STATS_KEYS | IGNORE_DEVICE_KEYS
 
+_IGNORE_KEYS_BY_MODEL_TYPE = {
+    #
+    # `lastMotion` from cameras update every 100 milliseconds when a motion event is active
+    # this overrides the behavior to only update `lastMotion` when a new event starts
+    #
+    ModelType.CAMERA: {"lastMotion"},
+    #
+    # `cameraIds` is updated every 10s, but we don't need to process it since bootstrap
+    # is resynced every so often anyways.
+    #
+    ModelType.CHIME: {"cameraIds"},
+}
+
+
+IGNORE_DEVICE_KEYS_BY_MODEL_TYPE = {
+    model_type: IGNORE_DEVICE_KEYS | keys
+    for model_type, keys in _IGNORE_KEYS_BY_MODEL_TYPE.items()
+}
+STATS_AND_IGNORE_DEVICE_KEYS_BY_MODEL_TYPE = {
+    model_type: STATS_AND_IGNORE_DEVICE_KEYS | keys
+    for model_type, keys in _IGNORE_KEYS_BY_MODEL_TYPE.items()
+}
+
+
 CAMERA_EVENT_ATTR_MAP: dict[EventType, tuple[str, str]] = {
     EventType.MOTION: ("last_motion", "last_motion_event_id"),
     EventType.SMART_DETECT: ("last_smart_detect", "last_smart_detect_event_id"),
@@ -77,35 +100,31 @@ CAMERA_EVENT_ATTR_MAP: dict[EventType, tuple[str, str]] = {
         "last_smart_audio_detect_event_id",
     ),
     EventType.RING: ("last_ring", "last_ring_event_id"),
+    EventType.NFC_CARD_SCANNED: (
+        "last_nfc_card_scanned",
+        "last_nfc_card_scanned_event_id",
+    ),
+    EventType.FINGERPRINT_IDENTIFIED: (
+        "last_fingerprint_identified",
+        "last_fingerprint_identified_event_id",
+    ),
 }
 
 
 def _process_light_event(event: Event, light: Light) -> None:
-    if _event_is_in_range(event, light.last_motion):
-        light.last_motion_event_id = event.id
-
-
-def _event_is_in_range(event: Event, dt: datetime | None) -> bool:
-    """Check if event is in range of datetime."""
-    return (
-        dt is None or event.start >= dt or (event.end is not None and event.end >= dt)
-    )
+    light.last_motion_event_id = event.id
 
 
 def _process_sensor_event(event: Event, sensor: Sensor) -> None:
     if event.type is EventType.MOTION_SENSOR:
-        if _event_is_in_range(event, sensor.motion_detected_at):
-            sensor.last_motion_event_id = event.id
+        sensor.last_motion_event_id = event.id
     elif event.type in {EventType.SENSOR_CLOSED, EventType.SENSOR_OPENED}:
-        if _event_is_in_range(event, sensor.open_status_changed_at):
-            sensor.last_contact_event_id = event.id
+        sensor.last_contact_event_id = event.id
     elif event.type is EventType.SENSOR_EXTREME_VALUE:
-        if _event_is_in_range(event, sensor.extreme_value_detected_at):
-            sensor.extreme_value_detected_at = event.end
-            sensor.last_value_event_id = event.id
+        sensor.extreme_value_detected_at = event.end
+        sensor.last_value_event_id = event.id
     elif event.type is EventType.SENSOR_ALARM:
-        if _event_is_in_range(event, sensor.alarm_triggered_at):
-            sensor.last_value_event_id = event.id
+        sensor.last_value_event_id = event.id
 
 
 _CAMERA_SMART_AND_LINE_EVENTS = {
@@ -118,10 +137,6 @@ _CAMERA_SMART_AUDIO_EVENT = EventType.SMART_AUDIO_DETECT
 def _process_camera_event(event: Event, camera: Camera) -> None:
     event_type = event.type
     dt_attr, event_attr = CAMERA_EVENT_ATTR_MAP[event_type]
-    dt: datetime | None = getattr(camera, dt_attr)
-    if not _event_is_in_range(event, dt):
-        return
-
     event_id = event.id
     event_start = event.start
 
@@ -154,10 +169,6 @@ class ProtectDeviceRef(ProtectBaseObject):
     id: str
 
 
-_ModelType_NVR_value = ModelType.NVR.value
-_ModelType_Event_value = ModelType.EVENT.value
-
-
 class Bootstrap(ProtectBaseObject):
     auth_user_id: str
     access_key: str
@@ -172,6 +183,8 @@ class Bootstrap(ProtectBaseObject):
     sensors: dict[str, Sensor]
     doorlocks: dict[str, Doorlock]
     chimes: dict[str, Chime]
+    aiports: dict[str, AiPort]
+    ringtones: list[Ringtone]
     last_update_id: str
 
     # TODO:
@@ -179,6 +192,8 @@ class Bootstrap(ProtectBaseObject):
     # agreements
 
     # not directly from UniFi
+    keyrings: Keyrings = Keyrings()
+    ulp_users: UlpUsers = UlpUsers()
     events: dict[str, Event] = FixSizeOrderedDict()
     capture_ws_stats: bool = False
     mac_lookup: dict[str, ProtectDeviceRef] = {}
@@ -195,11 +210,20 @@ class Bootstrap(ProtectBaseObject):
         api: ProtectApiClient | None = data.get("api") or (
             cls._api if isinstance(cls, ProtectBaseObject) else None
         )
-        data["macLookup"] = {}
-        data["idLookup"] = {}
-        for model_type in ModelType.bootstrap_models_types_set():
-            key = model_type.devices_key
+        mac_lookup: dict[str, dict[str, str | ModelType]] = {}
+        id_lookup: dict[str, dict[str, str | ModelType]] = {}
+        data["idLookup"] = id_lookup
+        data["macLookup"] = mac_lookup
+
+        for model_type in ModelType.bootstrap_models_types_set:
+            key = model_type.devices_key  # type: ignore[attr-defined]
             items: dict[str, ProtectModel] = {}
+            if key not in data:
+                data[key] = {}
+                _LOGGER.error(
+                    f"Missing key in bootstrap: {key}. This may be fixed by updating Protect."
+                )
+                continue
             for item in data[key]:
                 if (
                     api is not None
@@ -208,12 +232,13 @@ class Bootstrap(ProtectBaseObject):
                 ):
                     continue
 
-                ref = {"model": model_type, "id": item["id"]}
-                items[item["id"]] = item
-                data["idLookup"][item["id"]] = ref
+                id_: str = item["id"]
+                ref = {"model": model_type, "id": id_}
+                items[id_] = item
+                id_lookup[id_] = ref
                 if "mac" in item:
                     cleaned_mac = normalize_mac(item["mac"])
-                    data["macLookup"][cleaned_mac] = ref
+                    mac_lookup[cleaned_mac] = ref
             data[key] = items
 
         return super().unifi_dict_to_dict(data)
@@ -225,17 +250,9 @@ class Bootstrap(ProtectBaseObject):
     ) -> dict[str, Any]:
         data = super().unifi_dict(data=data, exclude=exclude)
 
-        if "events" in data:
-            del data["events"]
-        if "captureWsStats" in data:
-            del data["captureWsStats"]
-        if "macLookup" in data:
-            del data["macLookup"]
-        if "idLookup" in data:
-            del data["idLookup"]
-
-        for model_type in ModelType.bootstrap_models_types_set():
-            attr = model_type.devices_key
+        pop_dict_tuple(data, ("events", "captureWsStats", "macLookup", "idLookup"))
+        for model_type in ModelType.bootstrap_models_types_set:
+            attr = model_type.devices_key  # type: ignore[attr-defined]
             if attr in data and isinstance(data[attr], dict):
                 data[attr] = list(data[attr].values())
 
@@ -250,8 +267,7 @@ class Bootstrap(ProtectBaseObject):
 
     @property
     def auth_user(self) -> User:
-        user: User = self._api.bootstrap.users[self.auth_user_id]
-        return user
+        return self._api.bootstrap.users[self.auth_user_id]
 
     @property
     def has_doorbell(self) -> bool:
@@ -298,20 +314,20 @@ class Bootstrap(ProtectBaseObject):
 
     def get_device_from_mac(self, mac: str) -> ProtectAdoptableDeviceModel | None:
         """Retrieve a device from MAC address."""
-        ref = self.mac_lookup.get(normalize_mac(mac))
-        if ref is None:
-            return None
-
-        devices: dict[str, ProtectModelWithId] = getattr(self, ref.model.devices_key)
-        return cast(ProtectAdoptableDeviceModel, devices.get(ref.id))
+        return self._get_device_from_ref(self.mac_lookup.get(normalize_mac(mac)))
 
     def get_device_from_id(self, device_id: str) -> ProtectAdoptableDeviceModel | None:
         """Retrieve a device from device ID (without knowing model type)."""
-        ref = self.id_lookup.get(device_id)
+        return self._get_device_from_ref(self.id_lookup.get(device_id))
+
+    def _get_device_from_ref(
+        self, ref: ProtectDeviceRef | None
+    ) -> ProtectAdoptableDeviceModel | None:
         if ref is None:
             return None
-        devices: dict[str, ProtectModelWithId] = getattr(self, ref.model.devices_key)
-        return cast(ProtectAdoptableDeviceModel, devices.get(ref.id))
+        devices_key = ref.model.devices_key
+        devices: dict[str, ProtectAdoptableDeviceModel] = getattr(self, devices_key)
+        return devices[ref.id]
 
     def process_event(self, event: Event) -> None:
         event_type = event.type
@@ -324,28 +340,9 @@ class Bootstrap(ProtectBaseObject):
 
         self.events[event.id] = event
 
-    def _create_stat(
-        self,
-        packet: WSPacket,
-        keys_set: Iterable[str] | None,
-        filtered: bool,
-    ) -> None:
-        if self.capture_ws_stats:
-            self._ws_stats.append(
-                WSStat(
-                    model=packet.action_frame.data["modelKey"],
-                    action=packet.action_frame.data["action"],
-                    keys=list(packet.data_frame.data),
-                    keys_set=[] if keys_set is None else list(keys_set),
-                    size=len(packet.raw),
-                    filtered=filtered,
-                ),
-            )
-
     def _process_add_packet(
         self,
         model_type: ModelType,
-        packet: WSPacket,
         data: dict[str, Any],
     ) -> WSSubscriptionMessage | None:
         obj = create_from_unifi_dict(data, api=self._api, model_type=model_type)
@@ -357,51 +354,41 @@ class Bootstrap(ProtectBaseObject):
             if TYPE_CHECKING:
                 assert isinstance(obj, NVR)
             self.nvr = obj
-        elif model_type in ModelType.bootstrap_models_types_set():
+        elif model_type in ModelType.bootstrap_models_types_set:
             if TYPE_CHECKING:
                 assert isinstance(obj, ProtectAdoptableDeviceModel)
-                assert isinstance(obj.model, ModelType)
-            key = obj.model.devices_key
             if not self._api.ignore_unadopted or (
                 obj.is_adopted and not obj.is_adopted_by_other
             ):
-                getattr(self, key)[obj.id] = obj
-                ref = ProtectDeviceRef(model=obj.model, id=obj.id)
-                self.id_lookup[obj.id] = ref
+                id_ = obj.id
+                getattr(self, model_type.devices_key)[id_] = obj
+                ref = ProtectDeviceRef(model=model_type, id=id_)
+                self.id_lookup[id_] = ref
                 self.mac_lookup[normalize_mac(obj.mac)] = ref
         else:
             _LOGGER.debug("Unexpected bootstrap model type for add: %s", model_type)
             return None
 
-        updated = obj.dict()
-
-        self._create_stat(packet, updated, False)
-
         return WSSubscriptionMessage(
             action=WSAction.ADD,
             new_update_id=self.last_update_id,
-            changed_data=updated,
+            changed_data=obj.model_dump(),
             new_obj=obj,
         )
 
     def _process_remove_packet(
-        self, model_type: ModelType, packet: WSPacket
+        self, model_type: ModelType, action: dict[str, Any]
     ) -> WSSubscriptionMessage | None:
-        devices: dict[str, ProtectDeviceModel] | None = getattr(
-            self, model_type.devices_key, None
-        )
-
+        devices_key = model_type.devices_key
+        devices: dict[str, ProtectDeviceModel] | None = getattr(self, devices_key, None)
         if devices is None:
             return None
 
-        device_id: str = packet.action_frame.data["id"]
+        device_id: str = action["id"]
         self.id_lookup.pop(device_id, None)
-        device = devices.pop(device_id, None)
-        if device is None:
+        if (device := devices.pop(device_id, None)) is None:
             return None
         self.mac_lookup.pop(normalize_mac(device.mac), None)
-
-        self._create_stat(packet, None, False)
         return WSSubscriptionMessage(
             action=WSAction.REMOVE,
             new_update_id=self.last_update_id,
@@ -409,36 +396,84 @@ class Bootstrap(ProtectBaseObject):
             old_obj=device,
         )
 
+    def _process_ws_keyring_or_ulp_user_message(
+        self,
+        action: dict[str, Any],
+        data: dict[str, Any],
+        model_type: ModelType,
+    ) -> WSSubscriptionMessage | None:
+        action_id = action["id"]
+        obj_from_bootstrap: UlpUserKeyringBase[ProtectModelWithId] = getattr(
+            self, to_snake_case(model_type.devices_key)
+        )
+        action_type = action["action"]
+        if action_type == "add":
+            add_obj = create_from_unifi_dict(data, api=self._api, model_type=model_type)
+            if TYPE_CHECKING:
+                model_class = MODEL_TO_CLASS.get(model_type)
+                assert model_class is not None and isinstance(add_obj, model_class)
+            add_obj = cast(ProtectModelWithId, add_obj)
+            obj_from_bootstrap.add(add_obj)
+            return WSSubscriptionMessage(
+                action=WSAction.ADD,
+                new_update_id=self.last_update_id,
+                changed_data=add_obj.model_dump(),
+                new_obj=add_obj,
+            )
+        if action_type == "remove":
+            to_remove = obj_from_bootstrap.by_id(action_id)
+            if to_remove is None:
+                return None
+            obj_from_bootstrap.remove(to_remove)
+            return WSSubscriptionMessage(
+                action=WSAction.REMOVE,
+                new_update_id=self.last_update_id,
+                changed_data={},
+                old_obj=to_remove,
+            )
+        if action_type == "update":
+            updated_obj = obj_from_bootstrap.by_id(action_id)
+            if updated_obj is None:
+                return None
+
+            old_obj = updated_obj.model_copy()
+            updated_data = {to_snake_case(k): v for k, v in data.items()}
+            updated_obj.update_from_dict(updated_data)
+
+            return WSSubscriptionMessage(
+                action=WSAction.UPDATE,
+                new_update_id=self.last_update_id,
+                changed_data=updated_data,
+                new_obj=updated_obj,
+                old_obj=old_obj,
+            )
+        _LOGGER.debug("Unexpected ws action for %s: %s", model_type, action_type)
+        return None
+
     def _process_nvr_update(
         self,
-        packet: WSPacket,
+        action: dict[str, Any],
         data: dict[str, Any],
         ignore_stats: bool,
     ) -> WSSubscriptionMessage | None:
         if ignore_stats:
-            for key in STATS_KEYS.intersection(data):
-                del data[key]
+            pop_dict_set(data, STATS_KEYS)
         # nothing left to process
         if not data:
-            self._create_stat(packet, None, True)
             return None
 
         # for another NVR in stack
-        nvr_id = packet.action_frame.data.get("id")
+        nvr_id: str | None = action.get("id")
         if nvr_id and nvr_id != self.nvr.id:
-            self._create_stat(packet, None, True)
             return None
 
-        data = self.nvr.unifi_dict_to_dict(data)
         # nothing left to process
-        if not data:
-            self._create_stat(packet, None, True)
+        if not (data := self.nvr.unifi_dict_to_dict(data)):
             return None
 
-        old_nvr = self.nvr.copy()
-        self.nvr = self.nvr.update_from_dict(deepcopy(data))
+        old_nvr = self.nvr.model_copy()
+        self.nvr = self.nvr.update_from_dict(data)
 
-        self._create_stat(packet, data, False)
         return WSSubscriptionMessage(
             action=WSAction.UPDATE,
             new_update_id=self.last_update_id,
@@ -450,23 +485,31 @@ class Bootstrap(ProtectBaseObject):
     def _process_device_update(
         self,
         model_type: ModelType,
-        packet: WSPacket,
         action: dict[str, Any],
         data: dict[str, Any],
         ignore_stats: bool,
+        is_ping_back: bool,
     ) -> WSSubscriptionMessage | None:
-        remove_keys = (
-            STATS_AND_IGNORE_DEVICE_KEYS if ignore_stats else IGNORE_DEVICE_KEYS
-        )
-        for key in remove_keys.intersection(data):
-            del data[key]
-        # `last_motion` from cameras update every 100 milliseconds when a motion event is active
-        # this overrides the behavior to only update `last_motion` when a new event starts
-        if model_type is ModelType.CAMERA and "lastMotion" in data:
-            del data["lastMotion"]
+        """
+        Process a device update packet.
+
+        If is_ping_back is True, the packet is an empty packet
+        that was generated internally as a result of an event
+        that will expire and result in a state change.
+        """
+        if ignore_stats:
+            remove_keys = STATS_AND_IGNORE_DEVICE_KEYS_BY_MODEL_TYPE.get(
+                model_type, STATS_AND_IGNORE_DEVICE_KEYS
+            )
+        else:
+            remove_keys = IGNORE_DEVICE_KEYS_BY_MODEL_TYPE.get(
+                model_type, IGNORE_DEVICE_KEYS
+            )
+
+        pop_dict_set(data, remove_keys)
+
         # nothing left to process
-        if not data:
-            self._create_stat(packet, None, True)
+        if not data and not is_ping_back:
             return None
 
         devices: dict[str, ProtectModelWithId] = getattr(self, model_type.devices_key)
@@ -474,37 +517,32 @@ class Bootstrap(ProtectBaseObject):
         if action_id not in devices:
             # ignore updates to events that phase out
             if model_type is not ModelType.EVENT:
-                _LOGGER.debug("Unexpected %s: %s", key, action_id)
+                _LOGGER.debug("Unexpected %s: %s", model_type, action_id)
             return None
 
         obj = devices[action_id]
         data = obj.unifi_dict_to_dict(data)
-        old_obj = obj.copy()
-        obj = obj.update_from_dict(deepcopy(data))
+
+        if not data and not is_ping_back:
+            # nothing left to process
+            return None
+
+        old_obj = obj.model_copy()
+        obj = obj.update_from_dict(data)
 
         if model_type is ModelType.EVENT:
             if TYPE_CHECKING:
                 assert isinstance(obj, Event)
             self.process_event(obj)
-        elif model_type is ModelType.CAMERA:
-            if TYPE_CHECKING:
-                assert isinstance(obj, Camera)
-            if "last_ring" in data and obj.last_ring:
-                is_recent = obj.last_ring + RECENT_EVENT_MAX >= utc_now()
-                _LOGGER.debug("last_ring for %s (%s)", obj.id, is_recent)
-                if is_recent:
-                    obj.set_ring_timeout()
         elif model_type is ModelType.SENSOR:
             if TYPE_CHECKING:
                 assert isinstance(obj, Sensor)
-            if "alarm_triggered_at" in data and obj.alarm_triggered_at:
-                is_recent = obj.alarm_triggered_at + RECENT_EVENT_MAX >= utc_now()
-                _LOGGER.debug("alarm_triggered_at for %s (%s)", obj.id, is_recent)
-                if is_recent:
+            if "alarm_triggered_at" in data and (trigged_at := obj.alarm_triggered_at):
+                if is_recent := trigged_at + RECENT_EVENT_MAX >= utc_now():
                     obj.set_alarm_timeout()
+                _LOGGER.debug("alarm_triggered_at for %s (%s)", obj.id, is_recent)
 
         devices[action_id] = obj
-        self._create_stat(packet, data, False)
         return WSSubscriptionMessage(
             action=WSAction.UPDATE,
             new_update_id=self.last_update_id,
@@ -518,81 +556,108 @@ class Bootstrap(ProtectBaseObject):
         packet: WSPacket,
         models: set[ModelType] | None = None,
         ignore_stats: bool = False,
+        is_ping_back: bool = False,
     ) -> WSSubscriptionMessage | None:
         """Process a WS packet."""
+        capture_ws_stats = self.capture_ws_stats
         action = packet.action_frame.data
         data = packet.data_frame.data
-        if self.capture_ws_stats:
-            action = deepcopy(action)
-            data = deepcopy(data)
+        keys = list(data) if capture_ws_stats else None
 
-        new_update_id: str = action["newUpdateId"]
+        new_update_id: str | None = action["newUpdateId"]
         if new_update_id is not None:
             self.last_update_id = new_update_id
 
+        message = self._make_ws_packet_message(
+            action, data, models, ignore_stats, is_ping_back
+        )
+
+        if capture_ws_stats:
+            if TYPE_CHECKING:
+                assert keys is not None
+            self._ws_stats.append(
+                WSStat(
+                    model=action["modelKey"],
+                    action=action["action"],
+                    keys=keys,
+                    keys_set=[] if message is None else list(message.changed_data),
+                    size=len(packet.raw),
+                    filtered=message is None,
+                ),
+            )
+
+        return message
+
+    def _make_ws_packet_message(
+        self,
+        action: dict[str, Any],
+        data: dict[str, Any],
+        models: set[ModelType] | None,
+        ignore_stats: bool,
+        is_ping_back: bool,
+    ) -> WSSubscriptionMessage | None:
+        """Process a WS packet."""
         model_key: str = action["modelKey"]
-        if model_key not in ModelType.values_set():
+        if (model_type := ModelType.from_string(model_key)) is ModelType.UNKNOWN:
             _LOGGER.debug("Unknown model type: %s", model_key)
-            self._create_stat(packet, None, True)
             return None
 
-        model_type = ModelType.from_string(model_key)
         if models and model_type not in models:
-            self._create_stat(packet, None, True)
             return None
 
         action_action: str = action["action"]
-        if action_action == "remove":
-            return self._process_remove_packet(model_type, packet)
-
-        if not data:
-            self._create_stat(packet, None, True)
-            return None
 
         try:
+            if model_type in {ModelType.KEYRING, ModelType.ULP_USER}:
+                return self._process_ws_keyring_or_ulp_user_message(
+                    action, data, model_type
+                )
+            if action_action == "remove":
+                return self._process_remove_packet(model_type, action)
+            if not data and not is_ping_back:
+                return None
             if action_action == "add":
-                return self._process_add_packet(model_type, packet, data)
+                return self._process_add_packet(model_type, data)
             if action_action == "update":
                 if model_type is ModelType.NVR:
-                    return self._process_nvr_update(packet, data, ignore_stats)
-                if model_type in ModelType.bootstrap_models_types_and_event_set():
+                    return self._process_nvr_update(action, data, ignore_stats)
+                if model_type in ModelType.bootstrap_models_types_and_event_set:
                     return self._process_device_update(
-                        model_type,
-                        packet,
-                        action,
-                        data,
-                        ignore_stats,
+                        model_type, action, data, ignore_stats, is_ping_back
                     )
         except (ValidationError, ValueError) as err:
-            self._handle_ws_error(action, err)
+            self._handle_ws_error(action_action, model_type, action, err)
 
         _LOGGER.debug(
             "Unexpected bootstrap model type deviceadoptedfor update: %s", model_key
         )
-        self._create_stat(packet, None, True)
         return None
 
-    def _handle_ws_error(self, action: dict[str, Any], err: Exception) -> None:
+    def _handle_ws_error(
+        self,
+        action_action: str,
+        model_type: ModelType,
+        action: dict[str, Any],
+        err: Exception,
+    ) -> None:
         msg = ""
-        if action["modelKey"] == "event":
-            msg = f"Validation error processing event: {action['id']}. Ignoring event."
+        device_id: str = action["id"]
+        if model_type is ModelType.EVENT:
+            msg = f"Validation error processing event: {device_id}. Ignoring event."
         else:
-            try:
-                model_type = ModelType.from_string(action["modelKey"])
-                device_id: str = action["id"]
-                task = asyncio.create_task(self.refresh_device(model_type, device_id))
-                self._refresh_tasks.add(task)
-                task.add_done_callback(self._refresh_tasks.discard)
-            except (ValueError, IndexError):
-                msg = f"{action['action']} packet caused invalid state. Unable to refresh device."
-            else:
-                msg = f"{action['action']} packet caused invalid state. Refreshing device: {model_type} {device_id}"
+            task = asyncio.create_task(self.refresh_device(model_type, device_id))
+            self._refresh_tasks.add(task)
+            task.add_done_callback(self._refresh_tasks.discard)
+            msg = (
+                f"{action_action} packet caused invalid state. "
+                f"Refreshing device: {model_type} {device_id}"
+            )
         _LOGGER.debug("%s Error: %s", msg, err)
 
     async def refresh_device(self, model_type: ModelType, device_id: str) -> None:
         """Refresh a device in the bootstrap."""
         try:
-            if model_type == ModelType.NVR:
+            if model_type is ModelType.NVR:
                 device: ProtectModelWithId = await self._api.get_nvr()
             else:
                 device = await self._api.get_device(model_type, device_id)
@@ -600,7 +665,6 @@ class Bootstrap(ProtectBaseObject):
             ValidationError,
             TimeoutError,
             asyncio.TimeoutError,
-            asyncio.CancelledError,
             ClientError,
             ServerDisconnectedError,
         ):
@@ -610,12 +674,11 @@ class Bootstrap(ProtectBaseObject):
         if isinstance(device, NVR):
             self.nvr = device
         else:
-            devices: dict[str, ProtectModelWithId] = getattr(
-                self, model_type.devices_key
-            )
+            devices_key = model_type.devices_key
+            devices: dict[str, ProtectModelWithId] = getattr(self, devices_key)
             devices[device.id] = device
         _LOGGER.debug("Successfully refresh model: %s %s", model_type, device_id)
 
     async def get_is_prerelease(self) -> bool:
-        """Get if current version of Protect is a prerelease version."""
-        return await self.nvr.get_is_prerelease()
+        """[DEPRECATED] Always returns False. Will be removed after HA 2025.8.0."""
+        return False
